--
--
--
--
--

function string.index(str, i)
	return str:sub(i, i)
end

function phys()
	return game:getSystem("physics").world
end

function f(...)
	return string.format(...)
end

function hex2rgb(hex)
    hex = hex:gsub("#","")
    return tonumber("0x"..hex:sub(1,2)), tonumber("0x"..hex:sub(3,4)), tonumber("0x"..hex:sub(5,6))
end

function getClass(class)
	local r = {}
	local e = game:getEntities()

	for index, entity in pairs(e) do
		if entity.class == class then
			table.insert(r, entity)
		end
	end

	return r
end

function recursiveEnumerate(folder, fileTree)
	local lfs = love.filesystem
	local filesTable = lfs.getDirectoryItems(folder)
	fileTree = ""
	for i,v in ipairs(filesTable) do
		local file = folder.."/"..v
		if lfs.isFile(file) then
			fileTree = fileTree.."\n"..file
		elseif lfs.isDirectory(file) then
			fileTree = fileTree.."\n"..file.." (DIR)"
			fileTree = recursiveEnumerate(file, fileTree)
		end
	end
	return fileTree
end

function getRooms()
	local r = recursiveEnumerate("assets/rooms"):separate("\n")
	local roomList = {}
	for index, room in pairs(r) do
		table.insert(roomList, room:sub(14, -6))
	end

	return roomList
end

function pickRoom(func)
	local rooms = getRooms()
	local choices = {}
	for index, room in pairs(rooms) do
		table.insert(choices, {
			label = room,
			callback = func,
			room = room,
			onclick = function(self)
				self.choices[self.selectedChoice].callback(self.choices[self.selectedChoice].room)
			end,
		})
	end
	printf("there are %i rooms", #rooms)
	printf("there are %i choices", #choices)

	choose {
		text = "Choose your room.",
		buttons = choices,
		font = "system",
		choiceFont = "system",
	}
end

function chooseWarp()
	pickRoom(function(room)
		game.lastUsedWarp = f("%s#entrance", room)
		warp(f("%s#entrance", room))
	end)
end

function oppositeDir(dir)
	if dir == "left" then
		return "right"
	elseif dir == "right" then
		return "left"
	elseif dir == "up" then
		return "down"
	else
		return "up"
	end
end

function getNearest(list, x, y)
	local nearest = false

	for index, entity in pairs(list) do
		if nearest == false then
			nearest = entity
		else
			if math.dist(x, y, entity.x, entity.y) < math.dist(x, y, nearest.x, nearest.y) then
				nearest = entity
			end
		end
	end

	return nearest
end

function track(t, index)
	return t[math.clamp(index, 1, #t)]
end

function table.mix(...)
	local r = {}

	for i = 1, select("#", ...) do
		table.insert(r, t)
	end

	return r
end

function trajectory(drag, gravity, x, y, angle)
	local velocity = math.rotate(x, y, angle)

 	local high = {
		x = 0,
		y = velocity.y / gravity.y
	}

	local far = {
		x = velocity.x / drag.x,
		y = 200
	}

	return high, far
end

function getFormattedLength(str)
	local formatted = formatText(str)
	local unformatted = ''
	local length = 0

	for index, text in pairs(formatted) do
		if type(text) == "string" then
			length = length + #text
			formatted = formatted .. text
		end
	end

	return length, unformatted
end

function systemWrite(file, data)
	if not (os and io) then
	    notice "Either the OS or IO library is missing. Saving to appdata."
	    love.filesystem.write(file, data)
	    printf("Saved to %s", file)
	else
	    print "Saving with OS/IO"
	    local obj = io.open(f("%s/%s", love.filesystem.getWorkingDirectory( ), file), "w")
	    io.output(obj)
	    io.write(data)
	    io.close(obj)
		print "Done saving with OS/IO."
	end
end

function tableFormat(text, values)
	local formatted = {}
	local ignoreChars = 0

	for i = 1, #text do
		if ignoreChars > 0 then
			ignoreChars = ignoreChars - 1
		else
			if text:sub(i, i) == "%" then
				local isColor = false

				for name, color in pairs(values) do
					if text:sub(i, i + #name) == "%" .. name then
						ignoreChars = #name
						table.insert(formatted, tostring(color))
						isColor = true
					end
				end

				if not isColor then
					table.insert(formatted, text:sub(i, i))
				end
			else
				table.insert(formatted, text:sub(i, i))
			end
		end
	end

	return table.concat(formatted, "")
end

function outline_print(text, x, y, bg, fg)
	local r, g, b, a = love.graphics.getColor()

	bg = bg or {0, 0, 0}
	fg = fg or {255, 255, 255}
	love.graphics.setColor(bg)
	for index, offset in pairs(game.definitions.directions) do
		love.graphics.print(text, x + offset.x, y + offset.y)
	end

	love.graphics.setColor(fg)
	love.graphics.print(text, x, y)
	love.graphics.setColor({r, g, b, a})
end

function json(file)
	return game.json.decode(love.filesystem.read(file))
end

function getMapTile(map, x, y)
	if type(map) == "string" then
		map = game:get(map)
	end

	if not map then
		return false
	end

	if x > map.map.width or x < 0
	or y > map.map.height or y < 0 then
		return false
	end

	return map.map.tiles[y][x]
end

function color(name)
	return game.palette[name]
end

function queueText(text)
	local t = {}

	for index, str in pairs(text:separate("\n")) do
		table.insert(t, formatText(str))
	end

	return t
end

function formatText(text)
	local formatted = {}
	local ignoreChars = 0

	for i = 1, #text do
		if ignoreChars > 0 then
			ignoreChars = ignoreChars - 1
		else
			if text:sub(i, i) == "^" then
				local isColor = false

				for name, color in pairs(game.palette) do
					if text:sub(i, i + #name) == "^" .. name then
						ignoreChars = #name
						table.insert(formatted, color)
						isColor = true
					end
				end

				if not isColor then
					table.insert(formatted, text:sub(i, i))
				end
			else
				table.insert(formatted, text:sub(i, i))
			end
		end
	end

	return formatted
end

function create_2d_array(width, height, fill, copy)
	local grid = {}

	for i = 1, height do
	    grid[i] = {}

	    for j = 1, width do
			if copy then
	        	grid[i][j] = deepcopy(fill)
			else
				grid[i][j] = fill
			end
	    end
	end

	return grid
end

function ui(id, value)
	local ui = game:get(id)

	if not ui then
		return nil
	end

	if not ui.interface then
		return nil
	end

	if ui.interface.type == "Label" then
		if value then
			ui.interface.state = value
		end
		return ui.interface.state
	elseif ui.interface.type == "Input" or ui.interface.type == "Textarea" then
		if value then
			ui.interface.state.text = tostring(value)
		end
		return ui.interface.state.text
	elseif ui.interface.type == "Checkbox" then
		if value ~= nil then
			ui.interface.state.checked = value
		end
		return ui.interface.state.checked
	elseif ui.interface.type == "Slider" then
		if value then
			ui.interface.state.value = tonumber(value)
		end
		return ui.interface.state.value
	elseif ui.interface.type == "List" then
		if value then
			ui.interface.state.entries = value
		end
		return ui.interface.state.entries
	elseif ui.interface.type == "Stepper" then
		if value then
			if type(value) == "table" then
				ui.interface.state.steps = value
			elseif type(value) == "number" then
				ui.interface.state.position = math.clamp(value, 1, #ui.interface.state.steps)
			elseif type(value) == "string" then
				local correctStep = false
				for index, step in ipairs(ui.interface.state.steps) do
					if step == value then
						correctStep = index
					end
				end

				if not correctStep then
					table.insert(ui.interface.state.steps, value)
					ui.interface.state.position = #ui.interface.state.steps
				else
					ui.interface.state.position = correctStep
				end
			end
		end
		return ui.interface.state.steps[ui.interface.state.position]
	end
end

function save_file(file, target)
	if target == nil then
		target = "assets"
	end
	local path, filename, ext = GetFileName(file:getFilename())
	local targetDir = love.filesystem.getWorkingDirectory() .. string.format("/%s/%s", target, filename)
	os.execute(string.format("cp \"%s\" \"%s\"", file:getFilename(), targetDir))

	local f = io.open(targetDir)
	if f ~= nil then
		io.close(f)
		return true
	else
		return false
	end
end

function setMapTile(map, x, y, id)
	if type(map) == "string" then
		map = game:get(map)
	end

	if not map then
		return false
	end

	if x > map.map.width or x < 0
	or y > map.map.height or y < 0 then
		return false
	end

	table.insert(map.map.changedTiles, {x = x, y = y, id = id})
end

function getDimensions(entity)
	local entity = game:get(entity)
	if entity then
		return entity.x, entity.y, entity.z, entity.width, entity.height
	else
		return false
	end
end

function compress(entity)
	local model = {}

	model.class = entity.class

	model.animation = {
		schema = entity.animation.schema,
		current = entity.animation.current
	}

	if entity.behavior then
		model.behavior = {
			name = entity.behavior.name
		}
	end

	if entity.stats then
		model.stats = {
			name = entity.stats.name,
			base = entity.stats.base,

			behavior = entity.stats.behavior,

			health = entity.stats.health,
			maxHealth = entity.stats.maxHealth,

			experience = entity.stats.experience,
			level = entity.stats.level,

			isEnemy = entity.stats.isEnemy,

			expDrop = entity.stats.expDrop,
			itemDrops = entity.stats.itemDrops,

			flavor = entity.stats.flavor,
			baseStats = entity.stats.baseStats,
			controllable = entity.stats.controllable,
			verb = entity.stats.verb,
			skills = entity.stats.skills,

			physics = {
				width = entity.stats.physics.width,
				height = entity.stats.physics.height,
				origin = {
					x = entity.stats.physics.origin.x,
					y = entity.stats.physics.origin.y,
				}
			},

			sprite = {
				sprite = entity.stats.sprite.sprite,
				origin = {
					x = entity.stats.sprite.origin.x,
					y = entity.stats.sprite.origin.y,
				},
				flip = entity.stats.sprite.flip,
			}
		}
	end

	model.sprite = {
		sprite = entity.sprite.sprite,
		layer = entity.sprite.layer,
		anchored = entity.sprite.anchored,
		origin = {
			x = entity.sprite.origin.x,
			y = entity.sprite.origin.y,
		},
		scale = {
			x = entity.sprite.scale.x,
			y = entity.sprite.scale.y
		},
		offset = {
			x = entity.sprite.offset.x,
			y = entity.sprite.offset.y
		},
		color = {
			r = entity.sprite.color.r,
			g = entity.sprite.color.g,
			b = entity.sprite.color.b,
			a = entity.sprite.color.a,
		}
	}

	model.physics = {
		width = entity.physics.width,
		height = entity.physics.height,
		velocity = {
			x = entity.physics.velocity.x,
			y = entity.physics.velocity.y,
			max = {
				x = entity.physics.velocity.max.x,
				y = entity.physics.velocity.max.y
			}
		},
		offset = {
			x = entity.physics.offset.x,
			y = entity.physics.offset.y,
		},
		drag = {
			x = entity.physics.drag.x,
			y = entity.physics.drag.y
		},
		gravity = {
			x = entity.physics.gravity.x,
			y = entity.physics.gravity.y
		},
		collides = entity.physics.collides
	}

	print(table.inspect(model))
	return model
end

function warp_to(room, target)
	printf("Warping to room %s, target %s", room, target)
	game.mapMode = "warp"

	if not love.filesystem.exists(f("assets/rooms/%s.json", room)) then
		room = "construction"
	end


	r = load_room(room)

	local _target = game:get(target)

	if not _target then
		r = load_room("construction")
		_target = game:get("return_warp")
		game:get("return_warp_b").target = f("%s#%s", game.lastRoom, game.lastWarp)
	end

	if not game:get("player") then
		game:add("player")
	end

	local player = game:get("player")
	game.lastWarp = target
	game.lastRoom = room

	if r.properties.script then
		script(r.properties.script)
	end

	player.x = math.floor(_target.x + (_target.physics.width / 2) - (player.physics.width / 2))
	player.y = math.floor(_target.y + (_target.physics.height / 2) - (player.physics.height / 2))
	player.x = player.x - player.physics.offset.x
	player.y = player.y - player.physics.offset.y
end

function load_room(r)
	printf("Loading room %s", r)
	script "preload_room"

	roomWidth = 0
	roomHeight = 0

	game:getSystem("physics").world = game.physics.newWorld(32)

	for index, entity in pairs(game.scenes[game.activeScene].entities) do
		if not entity.isKey then
			game:remove(entity)
		end
	end

	game:cleanup()

	local zIndex = 0

	if not love.filesystem.exists(f("assets/rooms/%s.json", r)) then
		error(f("Room doesn't exist: %s", r))
	end

	if json(f("assets/rooms/%s.json", r)).version < 1.2 then
		map = require("game.map_old")(r)
	else
		map = game:map()
		map:load(r)
		map:scenify()
	end

	game.roomWidth = map.pixelWidth
	game.roomHeight = map.pixelHeight
	game:scene().roomWidth = game.roomWidth
	game:scene().roomHeight = game.roomHeight

	game.mapData = map

	script "postload_room"
	return map
end

function printf(...)
	print(string.format(...))
end

function math.dist(x1,y1, x2,y2) return ((x2-x1)^2+(y2-y1)^2)^0.5 end
function math.lerp(a,b,t) return (1-t)*a + t*b end
function math.rotate(x, y, angle)
  local c = math.cos(angle)
  local s = math.sin(angle)
  x = c * x - s * y
  y = s * x + c * y
  return {x = x, y = y}
end
function reduce_path(path)
	-- The path.
	local path = path:separate('/')

	for index, node in ipairs(path) do
		for i = 1, #node do
			if node:sub(i, i + 1) == '.' and index ~= #path then
				table.remove(path, index)
			end
		end
	end

	local returnPath = ''

	for index, node in pairs(path) do
		returnPath = returnPath .. '/' .. node
	end

	return returnPath
end

function math.rsign()
	return math.random(0, 1) == 1 and 1 or -1
end

function table.recurse(t, dir, value)
	local state = false
	if value ~= nil then
		state = {}
	else
		state = t
	end

	local stateCheck = state

	for index, key in ipairs(dir) do
		if index == #dir then
			if value == nil then
				return stateCheck[key]
			else
				stateCheck[key] = value
			end
		else
			if value ~= nil then
				stateCheck[key] = {}
			end

			stateCheck = stateCheck[key]
		end
	end

	return table.merge(t, state)
end

function string.separate(inputstr, sep)
	if sep == nil then
			sep = "%s"
	end

	local t={} ; i=1

	for str in string.gmatch(inputstr, "([^"..sep.."]+)") do
			t[i] = str
			i = i + 1
	end

	return t
end

function math.cerp(a,b,t)
	local f=(1-math.cos(t*math.pi))*.5
	return a*(1-f)+b*f
end

function deepcopy(o, seen)
	seen = seen or {}
	if o == nil then return nil end
	if seen[o] then return seen[o] end

	local no

	if type(o) == 'table' then
		no = {}
	    seen[o] = no

	    for k, v in next, o, nil do
	    	no[deepcopy(k, seen)] = deepcopy(v, seen)
	    end
	    setmetatable(no, deepcopy(getmetatable(o), seen))
	else -- number, string, boolean, etc
	    no = o
	end

	return no
end

function math.clamp(val, lower, upper, loop)
	if(loop == nil) then
		loop = false
	end

    assert(val and lower and upper, "math.clamp called with not enough arguments.")
    if lower > upper then lower, upper = upper, lower end -- swap if boundaries supplied the wrong way

	if(loop) then
		if(val < lower) then
			return upper
		elseif(val > upper) then
			return lower
		else
			return val
		end
	end

    return math.max(lower, math.min(upper, val))
end

function table.merge(t1, t2)
    for k,v in pairs(t2) do
        if type(v) == "table" then
            if type(t1[k] or false) == "table" then
                table.merge(t1[k] or {}, t2[k] or {})
            else
                t1[k] = v
            end
        else
            t1[k] = v
        end
    end
    return t1
end

function table.contains(table, value)
	local hasValue = false

	for key, _value in pairs(table) do
		if(_value == value) then
			hasValue = true
		end
	end

	return hasValue
end

function string.slice(str, max_line_length)
   local lines = {}
   local line
   str:gsub('(%s*)(%S+)',
      function(spc, word)
         if not line or #line + #spc + #word > max_line_length then
            table.insert(lines, line)
            line = word
         else
            line = line..spc..word
         end
      end
   )
   table.insert(lines, line)
   return lines
end

function string.capitalize(str)
    return (str:gsub("^%l", string.upper))
end

function math.sign(number)
	if(number > 0) then
		return 1
	elseif(number < 0) then
		return -1
	else
		return 0
	end
end

function GetFileName(url)
  return url:match "(.-)([^\\/]-%.?([^%.\\/]*))$"
end

function GetFileExtension(url)
  return url:match("^.+(%..+)$")
end

function string.random(length)
	if(game._charset == nil) then
		game._charset = {}
		game._charsetLength = 0

		for i = 48, 57 do
			table.insert(game._charset, i)
			game._charsetLength = game._charsetLength + 1
		end

		for i = 65, 90 do
			table.insert(game._charset, string.char(i))
			game._charsetLength = game._charsetLength + 1
		end

		for i = 97, 122 do
			table.insert(game._charset, string.char(i))
			game._charsetLength = game._charsetLength + 1
		end

		print('Charset loaded')
	end

	local stringReturn = ''

	for i = 1, length do
		local random = love.math.random(1, game._charsetLength)
		stringReturn = stringReturn .. game._charset[random]
	end

	return stringReturn
end

function math.angle(x1,y1, x2,y2)
	return math.atan2(y2-y1, x2-x1)
end

function math.dist(x1,y1, x2,y2) return ((x2-x1)^2+(y2-y1)^2)^0.5 end

function math.rotate(x, y, angle)
	local cs = math.cos(angle)
	local sn = math.sin(angle)

	return {
		x = x * cs - y * sn,
		y = x * sn + y * cs
	}
end
