--
--
--
--
--

local game = {
	-- Libraries
	audio = require 'lib.sound',
	baton = require 'lib.baton',
	json = require 'lib.json',
	lovebird = require 'lib.lovebird',
	moonshine = require 'lib.moonshine',
	statemachine = require 'lib.statemachine',
	physics = require 'lib.bump',
	patchy = require 'lib.patchy',
	suit = require 'lib.suit',
	tween = require 'lib.tween',

	-- Internal structures
	assets = require 'game.assets',
	camera = require 'game.camera',
	system = require 'game.system',
	component = require 'game.component',
	scene = require 'game.scene',
	definitions = require 'game.definitions',
	promise = require 'game.promise',
	newQueue = require 'game.queue',
	event = require 'game.event',
	map = require 'game.map',
	cutscene = require 'game.cutscene'(),

	file = false,

	debugMode = false,
	debugModes = 2,
	production = false,

	width = 640,
	height = 480,
	scale = 2,
	screenshotScale = 3,
	title = "Team Slime Engine",
	icon = "assets/icon_slime.png",
	identity = "slime",
	version = "0.2.0a",

	scenes = {},
	sceneList = {"main"},
	activeScene = "main",
	queuedScene = false,

	queues = {},
	targetContext = "entity",
	queued = 0,

	scripts = {},
	room = "nah",
	rooms = {},

	background = {0, 0, 0},
	buffer = false,
	layers = {},
	layerOrder = {
		"bg_effect",
		"background",
		"map",
		"map_props",
		"entities",
		"fx",
		"particles",
		"map_fg",
		"suit",
		"ui_bg_b",
		"ui_bg",
		"ui",
		"ui_fg",
		"foreground",
		"debug"
	},
	layerParallax = {
		background = 0.75,
		foreground = 1.5,
	},
	defaultParallax = 1,

	tab = false,
	queuedTab = false,

	fonts = {},
	fontSpecs = {},
	fontList = {
		{
			name = "pixel",
			source = "assets/fonts/Compass.ttf",
			size = 21,
			height = 1.75,
		},
		{
			name = "dense",
			source = "assets/fonts/Compass.ttf",
			size = 21,
			height = 1.2,
		}
	},
	defaultFont = "npc",

	mouse = {},

	locked = false,
	blackout = false,

	windows = {},
	windowList = {
		dialogue = "assets/sprites/ui/dialog.png",
		system = "assets/sprites/ui/system.png",
		speech = "assets/sprites/ui/speech.png"
	},

	state = {
		combo = 0,
		health = 4,
	},
	speed = 1,
	mapMode = "room",

	encounters = {},

	speaker = false,
	useFilters = false,
	particles = {},

	runtime = 0,
}

function game:findQueue(object)

	if object.context == "game" then
		self.queue:add(object)
	elseif object.context == "scene" then
		self:scene().queue:add(object)
	elseif object.context == "entity" then
		if self.context then
			self.context.queue:add(object)
		end
	elseif object.context == "any" then
		object.context = self.targetContext
		self:findQueue(object)
	end
end

function game:machine(...)
	return self.statemachine.create(...)
end

function game:load()
	for index, scene in pairs(self.sceneList) do
		self.scenes[scene] = require(string.format('game.scenes.%s', scene))()
	end

	self = table.merge(self, self.json.decode(love.filesystem.read("assets/game.json")))

	love.window.setMode(self.scale * self.width, self.scale * self.height)
	love.graphics.setLineStyle "rough"
	self.buffer = love.graphics.newCanvas(self.scale * self.width, self.scale * self.height)
	love.window.setTitle(self.title)

	for index, layer in pairs(self.layerOrder) do
		self.layers[layer] = love.graphics.newCanvas(self.width, self.height)
		self.layers[layer]:setFilter("nearest", "nearest")
	end

	for index, font in pairs(self.fontList) do
		self.fonts[font.name] = love.graphics.newFont(font.font or font.source, font.size)
		self.fonts[font.name]:setLineHeight(font.height)

		self.fontSpecs[font.name] = font
	end
	love.graphics.setFont(self.fonts[self.defaultFont])

	for name, window in pairs(self.windowList) do
		self.windows[name] = self.patchy.load(window)
	end

	love.window.setIcon(love.image.newImageData(self.icon))

	love.filesystem.setIdentity(self.identity)

	game.particles = love.graphics.newParticleSystem(love.graphics.newImage("assets/sprites/particles/1x1.png"), 5000)
	game.particles:setParticleLifetime(0.1, 0.5)
	game.particles:setLinearAcceleration(-15, -15, 15, 15)
	game.particles:setEmissionRate(0)
	game.particles:start()

	--
	--
	--
	print "Rolling duds."

	for i = 1, 1000000 do
		math.random(1, 1000)
		love.math.random(1, 1000)
	end

	if love.filesystem.exists("save.json") then
		print "loading save"
		self.save = self.json.decode(love.filesystem.read("save.json"))
	else
		print "loading default save"
		self.save = self.json.decode(love.filesystem.read("assets/save.json"))
	end

	self.suit.theme.color = {
	    normal  = {bg = { 66, 66, 66}, fg = {188,188,188}},
	    hovered = {bg = { 50,153,187}, fg = {255,255,255}},
	    active  = {bg = {255,153,  0}, fg = {225,225,225}}
	}

	local controls = self.json.decode(love.filesystem.read(love.filesystem.exists("controls.json") and "controls.json" or "assets/controls.json"))
  	controls.joystick = love.joystick.getJoysticks()[1]

	game.shine = game.moonshine(game.moonshine.effects.crt)
			.chain(game.moonshine.effects.scanlines)

	local palette = {}
	local color = {}

	for line in love.filesystem.lines "assets/text/palette.txt" do
		local bites = line:separate(' ')
		local reduced = {}

		for index, bite in ipairs(bites) do
			if #bite > 3 then
				local pieces = bite:separate('\t')
				table.insert(reduced, pieces[1])
				table.insert(reduced, pieces[2])
			else
				table.insert(reduced, bite)
			end
		end

		print(table.inspect(reduced))
		palette[reduced[4]] = {
			reduced[1], reduced[2], reduced[3]
		}
		local bites = reduced[4]:separate('-')
		if bites[2] then
			print(table.inspect(bites))

			if not color[bites[1]] then
				color[bites[1]] = {}
			end

			color[bites[1]][5 - tonumber(bites[2])] = {tonumber(reduced[1]), tonumber(reduced[2]), tonumber(reduced[3])}
		end
	end

	self.palette = palette
	self.colors = color

	self.activePalette = false
	self.setPalette = function(self, color)
		if color == false then
			self.activePalette = false
			return false
		end

		local palette = self:getPalette(color)

		if palette then
			self.activePalette = color
			self.shine.dmg.palette = palette
		end
	end
	self.getPalette = function(self, name)
		return self.colors[name] or false
	end

	self.controls = self.baton.new(controls)
	self.queue = self:newQueue()


	self.audio:volume("all", self.save.settings.volume.master)
	self.audio:volume("sfx", self.save.settings.volume.sfx)
	self.audio:volume("music", self.save.settings.volume.music)
	self.audio:volume("ambient", self.save.settings.volume.ambient)


	script("load")
end

function game:saveGame()
	script "save"
	love.filesystem.write("save.json", self.json.encode(self.save))
	print "Saved game"
end

function game:new(...)
	return self.scenes[self.activeScene]:validate(...)
end

function game:add(...)
	return self.scenes[self.activeScene]:add(...)
end
function game:template(...)
	return self.scenes[self.activeScene]:template(...)
end
function game:get(...)
	return self.scenes[self.activeScene]:get(...)
end
function game:remove(...)
	return self.scenes[self.activeScene]:remove(...)
end
function game:getSystem(sys)
	return self.scenes[self.activeScene].systems[sys]
end

function game:getEntities(filter)
	return self.scenes[self.activeScene]:getEntities(filter)
end

function game:update(dt, speed)
	self.runtime = self.runtime + dt

	if self.queuedTab then
		self.tab = self.queuedTab
		self.queuedTab = false
	end

	if self.queuedScene then
		self.activeScene = self.queuedScene
		self.queuedScene = false
	end

	if self.locked then
		speed = 0
	end

	self.mouse = {
		real = {
			x = love.mouse.getX(),
			y = love.mouse.getY()
		},

		world = {
			x = math.floor(love.mouse.getX() / self.scale) + self.camera.x,
			y = math.floor(love.mouse.getY() / self.scale) + self.camera.y
		},

		x = math.floor(love.mouse.getX() / self.scale),
		y = math.floor(love.mouse.getY() / self.scale),
		[1] = love.mouse.isDown(1),
		[2] = love.mouse.isDown(2),
		pressed = self.mouse and (not self.mouse[1] and love.mouse.isDown(1)) or false,
	}

	local avgFPS = 1 / love.timer.getAverageDelta()

	love.window.setTitle(string.format("%s / FPS: %i", self.title, avgFPS))

	self.particles:update(dt * self.speed)


	self.controls:update()
	self.promise:update(dt, self.speed)

	if self.controls:pressed "toggle_debug" then
		self.debugMode = self.debugMode and self.debugMode + 1 or 1

		if self.debugMode > self.debugModes then
			self.debugMode = false
		end
	end

	if self.controls:pressed "fullscreen" then
		fullscreen()
	end

	if self.controls:pressed "screenshot" then
		screenshot()
	end

	if love.keyboard.isDown("a") then
		--battle {{ class = "wizard", x = 7, y = 2}}
	end

	if love.keyboard.isDown("`") then
		if not self:get("console") then
			self.locked = true
			local console = self:add("ui/console", {x = 0, y = game.height - 16})
			printf("Is active: %s", tostring(self.suit:isActive("console")))
		end
	end

	self.queued = 0
	for index, queue in pairs(self.queues) do
		self.queued = self.queued + #queue
	end

	self.audio:cleanup()

	self.queue:update(dt, self.speed)
	self.scenes[self.activeScene]:update(dt, self.speed)
	self.camera:update(dt, self.speed)
end

function game:setLayer(layer)
	local lc = love.graphics.getCanvas()
	love.graphics.setCanvas(self.layers[layer])

	return lc
end

function game:draw()
	self.scenes[self.activeScene]:draw()
end

function game:cleanup()
	self.scenes[self.activeScene]:cleanup()
end

return game
